/*
 *    Copyright  1996, 1997, 1998, 1999, 2000 by Remco Treffkorn
 *    Distributed under the terms of the Q Public License. Please see
 *    the included file LICENSE.
 */

#include "config.h"
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>

#include <X11/Xos.h>
#include <X11/cursorfont.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/FileSB.h>
#include <Xm/MainW.h>
#include <Xm/RowColumn.h>
#include <Xm/LabelG.h>
#include <Xm/Form.h>
#include <Xm/List.h>
#include <Xm/ScrollBar.h>
#include <X11/xpm.h>

#include "xmap.h"
#include "icons.h"
#include "splash.h"
#include "mode.h"
#include "debug.h"

extern Widget toplevel, toolbar;
extern Widget showToolBarButton();

XpmAttributes attr;
Pixmap smap;
Pixmap pmap;
Pixmap stipple;
Pixmap mappmap;
Pixmap dir_opened, dir_closed;

extern void select_map_dialog();
extern void track_toggle();
extern void do_legend_dialog();
extern void do_draw();
extern void do_help();
extern void do_vector();
extern void beep();
extern void do_back(), do_next(), do_previous(), do_hexit();
extern void save_objects();
extern void show_stamps();
extern void routen_cb();
extern void print_map();

extern void get_tiff_map();

static char stipple_bits[] =
{
    0x10,			// ...*....
     0x20,			// ..*.....
     0x40,			// .*......
     0x80,			// *.......
     0x01,			// .......*
     0x02,			// ......*.
     0x04,			// .....*..
     0x08,			// ....*...
};

static char stipple_bits1[] =
{
    0x18,			// ...**...
     0x24,			// ..*..*..
     0x42,			// .*....*.
     0x81,			// *......*
     0x41,			// .*.....*
     0x22,			// ..*...*.
     0x14,			// ...*.*..
     0x08,			// ....*...
};

static char stipple_bits2[] =
{
    0x55,			// .*.*.*.*
     0xaa,			// *.*.*.*.
     0x55,			// .*.*.*.*
     0xaa,			// *.*.*.*.
     0x55,			// .*.*.*.*
     0xaa,			// *.*.*.*.
     0x55,			// .*.*.*.*
     0xaa,			// *.*.*.*.
};

struct tbutton tbutton[] =
{
    {1, 0, open_xpm, "Open New Map", select_map_dialog},
    {1, 0, floppy_xpm, "Save", save_objects},
    {1, 0, printer_xpm, "Print Map", print_map},
    {1, 0, draw_xpm, "Draw", do_draw},
    {1, 0, stamp_xpm, "Stamp", show_stamps},
    {1, 0, route_xpm, "Create Route", routen_cb},
    {0, 0, measure_xpm, "Measure Distance/Bearing", beep},
    {0, 0, house_xpm, "Address"},
    {1, 0, key_xpm, "Show Legend", do_legend_dialog},
    {0, 0, list_xpm, "Route Directions"},
    {1, 0, overview_xpm, "Show Overview", beep},
    {0, 0, address_xpm, "Zip"},
    {0, 0, phone_xpm, "Phone"},
    {0, 0, what_xpm, "Place"},
    {1, 0, zoomin_xpm, "Zoom In", zoomin, &mapdata},
    {1, 0, zoomout_xpm, "Zoom Out", zoomout, &mapdata},
    {0, 0, zoomout4_xpm, "Zoom Out 4x", zoomout4, &mapdata},
    {1, 0, gps_xpm, "Toggle Tracking", track_toggle},
    {1, 0, help_xpm, "Help Topics", do_help},
    {0, 0, NULL}
};
#define TRACK_BUTTON 17


struct tbutton hbutton[] =
{
    {1, 0, larrow_xpm, "Back", do_back},
    {1, 0, darrow_xpm, "Next Chapter", do_next},
    {1, 0, uarrow_xpm, "Previous Chapter", do_previous},
    {1, 0, find_xpm, "Find", beep},
    {1, 0, exit_xpm, "Exit help", do_hexit},
    {0, 0, NULL},
};

struct features motion[] =
{
    {0, gps_xpm},
    {0, gps1_xpm},
    {0, gps2_xpm},
    {0, gps3_xpm},
    {0, gps4_xpm},
    {0, gps5_xpm},
    {0, NULL},
};
#define NFRAMES 6

struct features features[] =
{
    {0, rmajor_xpm},
    {0, rminor_xpm},
    {0, rnames_xpm},
    {0, rroute_xpm},
    {0, wmajor_xpm},
    {0, wminor_xpm},
    {0, wnames_xpm},
    {0, pmajor_xpm},
    {0, pminor_xpm},
    {0, orail_xpm},
    {0, obodys_xpm},
    {0, otints_xpm},
    {0, ozip_xpm},
    {0, NULL},
};

struct features pbutton[] =
{
    {0, bcont_xpm},
    {0, bsep_xpm},
    {0, brw_xpm},
    {0, bstop_xpm},
    {0, bplay_xpm},
    {0, bff_xpm},
    {0, NULL},
};

struct objects marks[] =
{
    {0, c_red_xpm},
    {0, c_yellow_xpm},
    {0, c_green_xpm},
    {0, bd_red_xpm},
    {0, bd_yellow_xpm},
    {0, bd_green_xpm},
    {0, reddot10_xpm},
    {0, NULL},
};

struct stamps objects[] =
{
    {redball_xpm},
    {grnball_xpm},
    {blueball_xpm},
    {redpin_xpm},
    {greenpin_xpm},
    {bluepin_xpm},
    {brnpin_xpm},
    {cynpin_xpm},
    {grypin_xpm},
    {lgrpin_xpm},
    {magpin_xpm},
    {orgpin_xpm},
    {purpin_xpm},
    {whtpin_xpm},
    {trnspin_xpm},
    {house_xpm},
    {NULL},
};

struct stamps d_arrow[] =
{
    {dir_nw_xpm},
    {dir_n_xpm},
    {dir_ne_xpm},
    {dir_w_xpm},
    {dir_exit_xpm},
    {dir_e_xpm},
    {dir_sw_xpm},
    {dir_s_xpm},
    {dir_se_xpm},
    {dis_nw_xpm},
    {dis_n_xpm},
    {dis_ne_xpm},
    {dis_w_xpm},
    {dir_exit_xpm},
    {dis_e_xpm},
    {dis_sw_xpm},
    {dis_s_xpm},
    {dis_se_xpm},
    {NULL},
};

Pixmap legend;
Pixmap splash;
Pixmap speedo100, speedo200, speedo999;
Pixmap bulls;
Pixmap sbulls;
Pixmap rband;

Pixmap dir_arrow[8];

Pixmap wp, wm;

Widget draww;
GC drawGC;
GC copyGC;
GC hlGC;
GC clearGC, uniGC;

Dimension width, height;


Pixmap get_darrow_pixmap(int obj)
{
    return d_arrow[obj].pmap;
}

Pixmap get_mark_pixmap(int obj)
{
    return marks[obj].pmap;
}

Pixmap get_mark_mask(int obj)
{
    return marks[obj].smap;
}

set_color(GC gc, String color)
{
    Pixel pixel;
    Display *dpy = XtDisplayOfObject(toplevel);

    pixel = get_pixel(toplevel, color);
    XSetForeground(dpy, gc, pixel);
}

set_font(GC gc, XFontStruct * font)
{
    XGCValues gcv;

    gcv.font = font->fid;
    XChangeGC(XtDisplay(draww), gc, GCFont, &gcv);
}

#define PIXMAPCURSOR

void set_cursor(int busy)
{
    XColor fg, bg;
    Cursor cursor;

    if (busy) {
	cursor = XCreateFontCursor(XtDisplay(draww), XC_watch);
    } else {
	fg.red = fg.blue = fg.green = 0;
	bg.red = bg.blue = bg.green = 0xffff;


	if (mode_is(MARK)) {
	    cursor = XCreateFontCursor(XtDisplay(draww), XC_pencil);
	} else if (mode_is(CROUTE)) {
	    cursor = XCreateFontCursor(XtDisplay(draww), XC_target);
	} else {
#ifndef PIXMAPCURSOR
	    cursor = XCreateFontCursor(XtDisplay(draww), XC_dotbox);
#else
	    cursor = XCreatePixmapCursor(XtDisplay(draww), sbulls, sbulls, &fg, &bg,
					 15, 15);
#endif
	}
    }
    XDefineCursor(XtDisplay(draww), XtWindow(draww), cursor);
}

void register_canvas(Widget w)
{
    XGCValues gcv;
    int n;
    Atom importList[1];
    Arg args[10];
    extern Atom PIXMAP_INDEX;
    extern void DropDraw();

    draww = w;

    gcv.foreground = BlackPixelOfScreen(XtScreen(draww));
    drawGC = XCreateGC(XtDisplay(draww),
		RootWindowOfScreen(XtScreen(draww)), GCForeground, &gcv);


    copyGC = XCreateGC(XtDisplay(draww),
		RootWindowOfScreen(XtScreen(draww)), GCForeground, &gcv);


    hlGC = XCreateGC(XtDisplay(draww),
		RootWindowOfScreen(XtScreen(draww)), GCForeground, &gcv);

    uniGC = XCreateGC(XtDisplay(draww),
		RootWindowOfScreen(XtScreen(draww)), GCForeground, &gcv);
    clearGC = XCreateGC(XtDisplay(draww),
		RootWindowOfScreen(XtScreen(draww)), GCForeground, &gcv);

    set_color(drawGC, "Blue1");


    n = 0;

#ifdef DND
    importList[0] = PIXMAP_INDEX;
    XtSetArg(args[n], XmNimportTargets, importList);
    n++;
    XtSetArg(args[n], XmNnumImportTargets, 1);
    n++;
    XtSetArg(args[n], XmNdropSiteOperations, XmDROP_COPY);
    n++;
    XtSetArg(args[n], XmNdropProc, DropDraw);
    n++;
    XmDropSiteRegister(draww, args, n);
#endif
}

void loadIcons(void)
{
    int i, b;
    Pixel fg, bg;

    Display *dpy = XtDisplayOfObject(toplevel);
    Screen *scr = XtScreenOfObject(toplevel);
    Window rootWindow = XRootWindowOfScreen(scr);
    XpmColorSymbol colorsymbols[1];

#if 1
    XtVaGetValues(toolbar,
		  XmNforeground, &fg,
		  XmNbackground, &bg,
		  NULL);

    /* translate the TRANSPARENT color 'None' into our background color */
    colorsymbols[0].name = "None";
    colorsymbols[0].value = NULL;
    colorsymbols[0].pixel = bg;
#endif


    attr.visual = DefaultVisual(dpy, DefaultScreen(dpy));
    attr.colormap = DefaultColormap(dpy, DefaultScreen(dpy));
    attr.depth = DefaultDepthOfScreen(scr);
    attr.exactColors = False;
    attr.closeness = 80000;

#if 1
    attr.colorsymbols = colorsymbols;
    attr.numsymbols = 1;
    attr.valuemask = XpmColorSymbols | XpmExactColors | XpmCloseness;
#else
    attr.valuemask = XpmExactColors | XpmCloseness;
#endif

    for (i = 0, b = 0; tbutton[i].data; i++) {
	if (XpmCreatePixmapFromData(XtDisplay(toplevel),
				    XtWindow(toplevel), tbutton[i].data,
				    &tbutton[i].pmap, NULL,
				    &attr) != XpmSuccess)
	    fprintf(stderr, "Error reading toolbar pixmap %d!\n", i);
	else if (tbutton[i].active)
	    tbutton[i].w = showToolBarButton(i, b++, tbutton[i].pmap, tbutton[i].func, tbutton[i].func_data);
    }
    for (i = 0, b = 0; hbutton[i].data; i++) {
	if (XpmCreatePixmapFromData(XtDisplay(toplevel),
				    XtWindow(toplevel), hbutton[i].data,
				    &hbutton[i].pmap, NULL,
				    &attr) != XpmSuccess)
	    fprintf(stderr, "Error reading help toolbar pixmap %d!\n", i);
	else if (tbutton[i].active)
	    hbutton[i].w = showHelpToolBarButton(i, b++, hbutton[i].pmap, hbutton[i].func, hbutton[i].func_data);
    }
    for (i = 0; features[i].data; i++) {
	if (XpmCreatePixmapFromData(XtDisplay(toplevel),
				    XtWindow(toplevel), features[i].data,
				    &features[i].pmap, NULL,
				    &attr) != XpmSuccess)
	    fprintf(stderr, "Error reading feature pixmap %d!\n", i);
    }
    for (i = 0; motion[i].data; i++) {
	if (XpmCreatePixmapFromData(XtDisplay(toplevel),
				    XtWindow(toplevel), motion[i].data,
				    &motion[i].pmap, NULL,
				    &attr) != XpmSuccess)
	    fprintf(stderr, "Error reading motion pixmap %d!\n", i);
    }
    for (i = 0; objects[i].data; i++) {
	if (XpmCreatePixmapFromData(XtDisplay(toplevel),
				    XtWindow(toplevel), objects[i].data,
				    &objects[i].pmap, NULL,
				    &attr) != XpmSuccess)
	    fprintf(stderr, "Error reading object pixmap %d!\n", i);
    }
    for (i = 0; pbutton[i].data; i++) {
	if (XpmCreatePixmapFromData(XtDisplay(toplevel),
				    XtWindow(toplevel), pbutton[i].data,
				    &pbutton[i].pmap, NULL,
				    &attr) != XpmSuccess)
	    fprintf(stderr, "Error reading pbutton pixmap %d!\n", i);
    }
    for (i = 0; d_arrow[i].data; i++) {
	if (XpmCreatePixmapFromData(XtDisplay(toplevel),
				    XtWindow(toplevel), d_arrow[i].data,
				    &d_arrow[i].pmap, NULL,
				    &attr) != XpmSuccess)
	    fprintf(stderr, "Error reading arrow pixmap %d!\n", i);
    }



    if (XpmCreatePixmapFromData(XtDisplay(toplevel),
				XtWindow(toplevel), legend_xpm,
				&legend, NULL,
				&attr) != XpmSuccess)
	fprintf(stderr, "Error reading legend pixmap!\n");

    if (XpmCreatePixmapFromData(XtDisplay(toplevel),
				XtWindow(toplevel), splash_xpm,
				&splash, NULL,
				&attr) != XpmSuccess)
	fprintf(stderr, "Error reading splash pixmap!\n");

    if (XpmCreatePixmapFromData(XtDisplay(toplevel),
				XtWindow(toplevel), speedo100_xpm,
				&speedo100, NULL,
				&attr) != XpmSuccess)
	fprintf(stderr, "Error reading speedo100 pixmap!\n");

    if (XpmCreatePixmapFromData(XtDisplay(toplevel),
				XtWindow(toplevel), speedo200_xpm,
				&speedo200, NULL,
				&attr) != XpmSuccess)
	fprintf(stderr, "Error reading speedo200 pixmap!\n");

    if (XpmCreatePixmapFromData(XtDisplay(toplevel),
				XtWindow(toplevel), speedo999_xpm,
				&speedo999, NULL,
				&attr) != XpmSuccess)
	fprintf(stderr, "Error reading speedo999 pixmap!\n");

    if (XpmCreatePixmapFromData(XtDisplay(toplevel),
				XtWindow(toplevel), dir_opened_xpm,
				&dir_opened, NULL,
				&attr) != XpmSuccess)
	fprintf(stderr, "Error reading dir_opened pixmap!\n");

    if (XpmCreatePixmapFromData(XtDisplay(toplevel),
				XtWindow(toplevel), dir_closed_xpm,
				&dir_closed, NULL,
				&attr) != XpmSuccess)
	fprintf(stderr, "Error reading dir_closed pixmap!\n");

    if (XpmCreatePixmapFromData(XtDisplay(toplevel),
				XtWindow(toplevel), map_xpm,
				&mappmap, NULL,
				&attr) != XpmSuccess)
	fprintf(stderr, "Error reading map pixmap!\n");

    if (XpmCreatePixmapFromData(XtDisplay(toplevel),
				XtWindow(toplevel), blueline_xpm,
				&rband, NULL,
				&attr) != XpmSuccess)
	fprintf(stderr, "Error reading rubber band pixmap!\n");


    // From here on we want shape masks.
    attr.valuemask = 0;

    for (i = 0; marks[i].data; i++) {
	if (XpmCreatePixmapFromData(XtDisplay(toplevel),
				    XtWindow(toplevel), marks[i].data,
				    &marks[i].pmap, &marks[i].smap,
				    &attr) != XpmSuccess)
	    fprintf(stderr, "Error reading mark pixmap %d!\n", i);
    }

    for (i = 0; objects[i].data; i++) {
	if (XpmCreatePixmapFromData(XtDisplay(toplevel),
				    XtWindow(toplevel), objects[i].data,
				    &pmap, &objects[i].smap,
				    &attr) != XpmSuccess)
	    fprintf(stderr, "Error reading object pixmap %d!\n", i);
    }
    if (i = XpmCreatePixmapFromData(XtDisplay(toplevel),
				    XtWindow(toplevel), bulls_xpm,
				    &bulls, &sbulls,
				    &attr) != XpmSuccess)
	fprintf(stderr, "Error(%d) reading bulls pixmap!\n", i);

    if (i = XpmCreatePixmapFromData(XtDisplay(toplevel),
				    XtWindow(toplevel), waypoint_xpm,
				    &wp, &wm,
				    &attr) != XpmSuccess)
	fprintf(stderr, "Error(%d) reading waypoint pixmap!\n", i);

    stipple = XCreateBitmapFromData(XtDisplay(toplevel), XtWindow(draww),
				    stipple_bits2, 8, 8);

}

char *get_tip(Widget w)
{
    int i;

    for (i = 0; tbutton[i].data; i++) {
	if (w == tbutton[i].w)
	    return tbutton[i].tip;
    }
    return " ??? ";
}

char *get_hlptip(Widget w)
{
    int i;

    for (i = 0; hbutton[i].data; i++) {
	if (w == hbutton[i].w)
	    return hbutton[i].tip;
    }
    return " ??? ";
}

Pixmap ReadMap(struct map * map)
{
    Pixmap back = (Pixmap) NULL;
    char bufp[256];

    attr.valuemask = XpmExactColors | XpmCloseness;

    snprintf(bufp, 255, "%s/%s.xpm", get_mapdir(), map->filename);
    if (XpmReadFileToPixmap(XtDisplay(toplevel), XtWindow(toplevel), bufp, &back, NULL, &attr) == XpmSuccess) {
	map->mwidth = attr.width;
	map->mheight = attr.height;
	if (map->pmap)
	    XFreePixmap(XtDisplay(toplevel), map->pmap);
	map->pmap = back;
	return (back);
    }
    snprintf(bufp, 255, "%s/%s.xpm.gz", get_mapdir(), map->filename);
    if (XpmReadFileToPixmap(XtDisplay(toplevel), XtWindow(toplevel), bufp, &back, NULL, &attr) == XpmSuccess) {
	map->mwidth = attr.width;
	map->mheight = attr.height;
	if (map->pmap)
	    XFreePixmap(XtDisplay(toplevel), map->pmap);
	map->pmap = back;
	return (back);
    }
    snprintf(bufp, 255, "%s/%s", get_mapdir(), map->filename);
    if (XpmReadFileToPixmap(XtDisplay(toplevel), XtWindow(toplevel), bufp, &back, NULL, &attr) == XpmSuccess) {
	map->mwidth = attr.width;
	map->mheight = attr.height;
	if (map->pmap)
	    XFreePixmap(XtDisplay(toplevel), map->pmap);
	map->pmap = back;
	return (back);
    }
    beep();
    return (back);
}

void update_map(char *name)
{
    char bufp[256];
    GetPrivateProfileString("Maps", "current", "", bufp, 255, inifile);
    if (strcmp(name, bufp) != 0) {
	GetPrivateProfileString("Maps", "map4", "", bufp, 255, inifile);
	if (*bufp)
	    WritePrivateProfileString("Maps", "map5", bufp, inifile);

	GetPrivateProfileString("Maps", "map3", "", bufp, 255, inifile);
	if (*bufp)
	    WritePrivateProfileString("Maps", "map4", bufp, inifile);

	GetPrivateProfileString("Maps", "map2", "", bufp, 255, inifile);
	if (*bufp)
	    WritePrivateProfileString("Maps", "map3", bufp, inifile);

	GetPrivateProfileString("Maps", "map1", "", bufp, 255, inifile);
	if (*bufp)
	    WritePrivateProfileString("Maps", "map2", bufp, inifile);

	GetPrivateProfileString("Maps", "current", "", bufp, 255, inifile);
	if (*bufp)
	    WritePrivateProfileString("Maps", "map1", bufp, inifile);

	WritePrivateProfileString("Maps", "current", name, inifile);

	WriteOutProfiles();
    }
}

void get_map(struct map *map)
{
    set_cursor(1);

    if (mapdata.type == VECTOR) {
	// Vector Map
	vector_map(map);
	do_vector(map);
	update_map(map->desname);
	cvll(mapdata.lat1, mapdata.lon1, mapdata.lat2, mapdata.lon2);
    } else if (mapdata.type == SHAPE) {
	shape_map(map);
	do_shape(map);
	update_map(map->desname);
	csll(mapdata.lat1, mapdata.lon1, mapdata.lat2, mapdata.lon2);
    } else if (mapdata.type == DRG) {
	init_map(map);
	do_drg(map);
	update_map(map->desname);
	draw_drg(0);
    } else if (mapdata.type == BSB) {
	do_bsb(map);
	update_map(map->desname);
	draw_bsb(0);
    } else if (ReadMap(map)) {
	init_map(map);
	update_map(map->desname);
	XCopyArea(XtDisplay(draww), map->pmap, XtWindow(draww), drawGC,
		  0, 0, width, height, 0, 0);
    }
    set_cursor(0);
}

void animate_button(int flag)
{
    static frame = 0;

    if (flag == 0) {
	frame = 0;
    } else {
	frame++;
	if (frame >= NFRAMES)
	    frame = 0;
    }
    if (tbutton[TRACK_BUTTON].w)
	XtVaSetValues(tbutton[TRACK_BUTTON].w,
		      XmNlabelPixmap, motion[frame].pmap,
		      NULL);
}


static int ox, oy;

void refresh(int warp)
{
    int x, y;

    if (mapdata.type == SRASTER) {
	XCopyArea(XtDisplay(draww), map.pmap, XtWindow(draww), drawGC,
	      mapdata.mx_offset, mapdata.my_offset, width, height, 0, 0);

	if (warp)
	    XWarpPointer(XtDisplay(draww), 0, 0, 0, 0, 0, 0,
			 ox - mapdata.mx_offset, oy - mapdata.my_offset);
    } else if (mapdata.type == DRG || mapdata.type == BSB) {
	XCopyArea(XtDisplay(draww), map.pmap, XtWindow(draww), drawGC,
		  0, 0, width, height, 0, 0);

	if (warp)
	    XWarpPointer(XtDisplay(draww), 0, 0, 0, 0, 0, 0,
			 (ox - mapdata.mx_offset) / mapdata.zoom,
			 (oy - mapdata.my_offset) / mapdata.zoom);
    } else {
	XCopyArea(XtDisplay(draww), map.pmap, XtWindow(draww), drawGC,
		  0, 0, width, height, 0, 0);
/*FIXME */
//      if (warp) XWarpPointer(XtDisplay(draww), 0, 0, 0, 0, 0, 0,
	//               ox - mapdata.mx_offset, oy - mapdata.my_offset);
    }

    redraw_mark(draww, NULL, NULL);
    redraw_objects();
    redraw_location();
}


void center(int x, int y)
{
    int refx, refy, offsetx, offsety;

    set_cursor(True);
    if (mapdata.type == VECTOR)
	vcenter(x, y);
    else {
	refx = width / 2;
	refy = height / 2;

	offsetx = x - refx;
	offsety = y - refy;

	ox = mapdata.mx_offset;
	oy = mapdata.my_offset;

	mapdata.mx_offset += offsetx * mapdata.zoom;
	mapdata.my_offset += offsety * mapdata.zoom;
	if (mapdata.mx_offset < 0)
	    mapdata.mx_offset = 0;
	if (mapdata.my_offset < 0)
	    mapdata.my_offset = 0;

	if (mapdata.mx_offset + width > map.mwidth)
	    mapdata.mx_offset = map.mwidth - width;
	if (mapdata.my_offset + height > map.mheight)
	    mapdata.my_offset = map.mheight - height;

	// Again, I know!
	if (mapdata.mx_offset < 0)
	    mapdata.mx_offset = 0;
	if (mapdata.my_offset < 0)
	    mapdata.my_offset = 0;

	if (mapdata.type == DRG)
	    drg_center();
	else if (mapdata.type == BSB)
	    bsb_center();
	else
	    refresh(1);

//      fprintf(stderr,"Center around %d,%d\nrefx=%d refy=%d\noffx=%d offy=%d\nxo=%d yo=%d\n",x,y,refx,refy,offsetx,offsety,mapdata.mx_offset, mapdata.my_offset);
    }
    set_cursor(False);
}

void change_raster(int x1, int y1, int x2, int y2)
{
    int dx, dy;
    int t, z;

    if (mapdata.zoom > 1) {

	dx = abs(x2 - x1);
	dy = abs(y2 - y1);

	t = MIN(width / dx, height / dy);
	if (!t)
	    t = 1;
	z = mapdata.zoom / t;
	if (!z)
	    z = 1;

	mapdata.mx_offset = ((MIN(x2, x1) + dx / 2) * mapdata.zoom + mapdata.mx_offset)
	    - (width / 2 * z);
	mapdata.my_offset = ((MIN(y2, y1) + dy / 2) * mapdata.zoom + mapdata.my_offset)
	    - (height / 2 * z);

	mapdata.zoom = z;

	if (mapdata.mx_offset < 0)
	    mapdata.mx_offset = 0;
	if (mapdata.my_offset < 0)
	    mapdata.my_offset = 0;

	if (mapdata.type == DRG)
	    draw_drg(0);
	else if (mapdata.type == BSB)
	    draw_bsb(0);
    }
}


static void draw_pixmap(Pixmap map, Pixmap mask, int x, int y, int offx, int offy, int sx, int sy)
{
    /* Set to dest x and y clip origin */
    XSetClipOrigin(XtDisplay(draww), drawGC, x - offx, y - offy);

    /* Set the clipping mask */
    XSetClipMask(XtDisplay(draww), drawGC, mask);

    XCopyArea(XtDisplay(draww), map,
	      XtWindow(draww), drawGC,
	      0, 0, sx, sy, x - offx, y - offy);
    XSetClipMask(XtDisplay(draww), drawGC, 0);
}

void draw_object_pixmap(int index, int x, int y)
{
    if (index < 0)
	draw_pixmap(wp, wm, x, y, 4, 4, 9, 9);
    else
	draw_pixmap(objects[index].pmap, objects[index].smap, x, y, 7, 7, 16, 16);
}

void graph_mark(status, x, y)
{
    if (status > 0 && status < 4) {

	/* Set to dest x and y clip origin */
	XSetClipOrigin(XtDisplay(draww), drawGC, x - 7, y - 7);

	/* Set the clipping mask */
	XSetClipMask(XtDisplay(draww), drawGC, marks[status - 1].smap);

	XCopyArea(XtDisplay(draww), marks[status - 1].pmap,
		  XtWindow(draww), drawGC,
		  0, 0, 16, 16, x - 7, y - 7);
	XSetClipMask(XtDisplay(draww), drawGC, 0);
    }
}

graph_unmark(x, y)
{
    int xo, yo;

    if (mapdata.type == SRASTER) {
	xo = mapdata.mx_offset;
	yo = mapdata.my_offset;
    } else {
	xo = 0;
	yo = 0;
    }
    XCopyArea(XtDisplay(draww), map.pmap,
	      XtWindow(draww), drawGC,
	      xo + x - 7, yo + y - 7,
	      16, 16, x - 7, y - 7);
}


static struct {
    int on;
    Position sx, sy;
    Position ex, ey;
    Dimension w, h;
} box;

void draw_box(Widget widget, int sx, int sy, int ex, int ey)
{
    int tmp;
    int w, h;

    if (ex < sx) {
	tmp = ex;
	ex = sx;
	sx = tmp;
    }
    if (ey < sy) {
	tmp = ey;
	ey = sy;
	sy = tmp;
    }
    w = ex - sx;
    h = ey - sy;


#ifdef ASPECTRATIOCORRECTION
    if (w > h) {
	h = height * w / width;
    } else {
	w = height * h / width;
    }
#endif

    box.sx = sx;
    box.ex = ex;
    box.sy = sy;
    box.ey = ey;
    box.w = w;
    box.h = h;

    XDrawRectangle(XtDisplay(widget), XtWindow(widget), drawGC,
		   sx, sy, w, h);
    box.on = True;
}

void undraw_box(Widget w)
{
    int xo, yo;

    if (mapdata.type == SRASTER) {
	xo = mapdata.mx_offset;
	yo = mapdata.my_offset;
    } else {
	xo = 0;
	yo = 0;
    }

    if (box.on) {
	XCopyArea(XtDisplay(w), map.pmap, XtWindow(w), drawGC,
		  xo + box.sx, yo + box.sy,
		  box.w, 1,
		  box.sx, box.sy);
	XCopyArea(XtDisplay(w), map.pmap, XtWindow(w), drawGC,
		  xo + box.sx, yo + box.sy,
		  1, box.h,
		  box.sx, box.sy);

	XCopyArea(XtDisplay(w), map.pmap, XtWindow(w), drawGC,
		  xo + box.sx, yo + box.ey,
		  box.w + 1, 1,
		  box.sx, box.ey);
	XCopyArea(XtDisplay(w), map.pmap, XtWindow(w), drawGC,
		  xo + box.ex, yo + box.sy,
		  1, box.h,
		  box.ex, box.sy);

	box.on = False;
    }
}

int is_box_on()
{
    return box.on;
}


static struct {
    int on;
    Position sx, sy;
    Position ex, ey;
} rubber;


void draw_rubber(Widget widget, int sx, int sy, int ex, int ey)
{
    XGCValues gcv;

    gcv.function = GXcopy;
    gcv.tile = rband;
    gcv.fill_style = FillTiled;
    gcv.ts_x_origin = 0;
    gcv.ts_y_origin = 0;
    gcv.line_width = 1;

    XChangeGC(XtDisplay(widget), copyGC,
	      GCFillStyle | GCFunction | GCTile | GCTileStipXOrigin |
	      GCTileStipYOrigin | GCLineWidth,
	      &gcv);

    rubber.sx = sx;
    rubber.ex = ex;
    rubber.sy = sy;
    rubber.ey = ey;
    XDrawLine(XtDisplay(widget), XtWindow(widget), copyGC, sx, sy, ex, ey);
    rubber.on = 1;
}

void undraw_rubber(Widget widget)
{
    XGCValues gcv;

    if (!rubber.on)
	return;

    gcv.function = GXcopy;
    gcv.tile = map.pmap;
    gcv.fill_style = FillTiled;
    if (mapdata.type == SRASTER) {
	gcv.ts_x_origin = -mapdata.mx_offset;
	gcv.ts_y_origin = -mapdata.my_offset;
    } else {
	gcv.ts_x_origin = 0;
	gcv.ts_y_origin = 0;
    }
    gcv.line_width = 2;

    XChangeGC(XtDisplay(widget), copyGC,
	      GCFillStyle | GCFunction | GCTile | GCTileStipXOrigin |
	      GCTileStipYOrigin | GCLineWidth,
	      &gcv);
    XDrawLine(XtDisplay(widget), XtWindow(widget), copyGC,
	      rubber.sx, rubber.sy, rubber.ex, rubber.ey);
    rubber.on = 0;
}

void highlight(Widget widget, XPoint * p, int n)
{
    XGCValues gcv;

    gcv.line_width = 16;
    gcv.function = GXcopy;
    gcv.stipple = stipple;
    gcv.fill_style = FillStippled;
    gcv.cap_style = CapRound;
    gcv.join_style = JoinRound;

    XChangeGC(XtDisplay(widget), hlGC,
	      GCFillStyle | GCFunction | GCStipple | GCLineWidth | GCCapStyle | GCJoinStyle,
	      &gcv);
    set_color(hlGC, "red");
    XDrawLines(XtDisplay(widget), XtWindow(widget), hlGC,
	       p, n, CoordModeOrigin);
}

void draw_route(Widget widget, XPoint * p, int n, int flag)
{
    XGCValues gcv;
    extern char *route_color();

    if (flag) {
	gcv.line_width = 2;
	gcv.function = GXcopy;
	gcv.fill_style = FillSolid;
	gcv.cap_style = CapRound;
	gcv.join_style = JoinRound;

	XChangeGC(XtDisplay(widget), hlGC,
	GCFillStyle | GCFunction | GCLineWidth | GCCapStyle | GCJoinStyle,
		  &gcv);

	if (strcmp("None", route_color())) {
	    set_color(hlGC, route_color());
	    XDrawLines(XtDisplay(widget), XtWindow(widget), hlGC,
		       p, n, CoordModeOrigin);
	}
	while (n--) {
	    draw_object_pixmap(-1, p[n].x, p[n].y);
	}
    } else {
#if 1
	gcv.line_width = 8;
	gcv.function = GXcopy;
	gcv.tile = map.pmap;
	gcv.fill_style = FillTiled;
	if (mapdata.type == SRASTER) {
	    gcv.ts_x_origin = -mapdata.mx_offset;
	    gcv.ts_y_origin = -mapdata.my_offset;
	} else {
	    gcv.ts_x_origin = 0;
	    gcv.ts_y_origin = 0;
	}
	XChangeGC(XtDisplay(widget), hlGC,
		  GCFillStyle | GCFunction | GCLineWidth | GCTile
		  | GCTileStipXOrigin | GCTileStipYOrigin,
		  &gcv);
	XDrawLines(XtDisplay(widget), XtWindow(widget), hlGC,
		   p, n, CoordModeOrigin);
#else
	gcv.line_width = 8;
	gcv.function = GXcopy;
	gcv.fill_style = FillSolid;
	gcv.cap_style = CapRound;
	gcv.join_style = JoinRound;

	XChangeGC(XtDisplay(widget), hlGC,
	GCFillStyle | GCFunction | GCLineWidth | GCCapStyle | GCJoinStyle,
		  &gcv);
	set_color(hlGC, "green");
	XDrawLines(XtDisplay(widget), XtWindow(widget), hlGC,
		   p, n, CoordModeOrigin);
#endif
    }
}


void mark_location(Position x, Position y)
{
    XGCValues gcv;
    extern char *route_color();

    gcv.line_width = 2;
    gcv.function = GXcopy;
    gcv.fill_style = FillSolid;
    gcv.cap_style = CapRound;
    gcv.join_style = JoinRound;

    XChangeGC(XtDisplay(draww), hlGC,
       GCFillStyle | GCFunction | GCLineWidth | GCCapStyle | GCJoinStyle,
	      &gcv);

    set_color(hlGC, "red");
    XDrawLine(XtDisplay(draww), XtWindow(draww), hlGC,
	      0, y, width, y);
    XDrawLine(XtDisplay(draww), XtWindow(draww), hlGC,
	      x, 0, x, height);
}

resize_cb(Widget w, XtPointer client_data, XmDrawingAreaCallbackStruct * cbs)
{
    extern struct map map;
    Dimension nwidth, nheight;
    int redraw = 0;

    XtVaGetValues(draww, XmNwidth, &nwidth, XmNheight, &nheight, NULL);

    if (nwidth > width || nheight > height)
	redraw++;

    width = nwidth;
    height = nheight;

    if (redraw) {
	switch (mapdata.type) {
	case VECTOR:
	    cvll(mapdata.lat1, mapdata.lon1, mapdata.lat2, mapdata.lon2);
	    break;

	case DRG:
	    draw_drg(0);
	    break;

	case BSB:
	    draw_bsb(0);
	    break;
	}
    }
}

redraw_cb(Widget w, XtPointer client_data, XmDrawingAreaCallbackStruct * cbs)
{
    extern struct map map;

    if (map.pmap) {
	if (mapdata.type == SRASTER)
	    XCopyArea(XtDisplay(draww), map.pmap, XtWindow(draww), drawGC,
		      cbs->event->xexpose.x + mapdata.mx_offset,
		      cbs->event->xexpose.y + mapdata.my_offset,
		   cbs->event->xexpose.width, cbs->event->xexpose.height,
		      cbs->event->xexpose.x, cbs->event->xexpose.y);
	else
	    XCopyArea(XtDisplay(draww), map.pmap, XtWindow(draww), drawGC,
		      cbs->event->xexpose.x,
		      cbs->event->xexpose.y,
		   cbs->event->xexpose.width, cbs->event->xexpose.height,
		      cbs->event->xexpose.x, cbs->event->xexpose.y);
    }
    redraw_mark(w, client_data, cbs);
    redraw_objects();
    redraw_location();
}
