#!/usr/bin/perl
# -*- perl -*-
use strict;

my (@dirlist) = @ARGV;
my ($mapdir, $firstchar);

sub lettonum($)
{
    my ($let) = @_;
    if (($let eq 'a') || ($let eq 'A'))
    {return 1;}
    if (($let eq 'b') || ($let eq 'B'))
    {return 2;}
    if (($let eq 'c') || ($let eq 'C'))
    {return 3;}
    if (($let eq 'd') || ($let eq 'D'))
    {return 4;}
    if (($let eq 'e') || ($let eq 'E'))
    {return 5;}
    if (($let eq 'f') || ($let eq 'F'))
    {return 6;}
    if (($let eq 'g') || ($let eq 'G'))
    {return 7;}
    if (($let eq 'h') || ($let eq 'H'))
    {return 8;}
}

sub nextlet($$)
{
    my ($nlat, $let) = @_;
    if ($let eq 'a')
    {return($nlat, 'b');}
    if ($let eq 'b')
    {return($nlat, 'c');}
    if ($let eq 'c')
    {return($nlat, 'd');}
    if ($let eq 'd')
    {return($nlat, 'e');}
    if ($let eq 'e')
    {return($nlat, 'f');}
    if ($let eq 'f')
    {return($nlat, 'g');}
    if ($let eq 'g')
    {return($nlat, 'h');}
    if ($let eq 'h')
    {$nlat++; return($nlat, 'a');}

    if ($let eq 'A')
    {return($nlat, 'B');}
    if ($let eq 'B')
    {return($nlat, 'C');}
    if ($let eq 'C')
    {return($nlat, 'D');}
    if ($let eq 'D')
    {return($nlat, 'E');}
    if ($let eq 'E')
    {return($nlat, 'F');}
    if ($let eq 'F')
    {return($nlat, 'G');}
    if ($let eq 'G')
    {return($nlat, 'H');}
    if ($let eq 'H')
    {$nlat++; return($nlat, 'A');}
}

sub nextnum($$)
{
    my ($nlon, $let) = @_;
    if ($let eq '1')
    {return($nlon, '2');}
    if ($let eq '2')
    {return($nlon, '3');}
    if ($let eq '3')
    {return($nlon, '4');}
    if ($let eq '4')
    {return($nlon, '5');}
    if ($let eq '5')
    {return($nlon, '6');}
    if ($let eq '6')
    {return($nlon, '7');}
    if ($let eq '7')
    {return($nlon, '8');}
    if ($let eq '8')
    {$nlon++; $nlon="0$nlon"; return($nlon, '1');}
}

sub lastnum($$)
{
    my ($nlon, $let) = @_;

    if ($let eq '1') {
	$nlon--;
	$nlon = "0$nlon";
        return($nlon, '8');
    }
    if ($let eq '2')
    {return($nlon, '1');}
    if ($let eq '3')
    {return($nlon, '2');}
    if ($let eq '4')
    {return($nlon, '3');}
    if ($let eq '5')
    {return($nlon, '4');}
    if ($let eq '6')
    {return($nlon, '5');}
    if ($let eq '7')
    {return($nlon, '6');}
    if ($let eq '8')
    {return($nlon, '7');}
}

sub lastlet($$)
{
    my ($nlat, $let) = @_;

    if ($let eq 'a')
    {$nlat--; return($nlat, 'h');}
    if ($let eq 'b')
    {return($nlat, 'a');}
    if ($let eq 'c')
    {return($nlat, 'b');}
    if ($let eq 'd')
    {return($nlat, 'c');}
    if ($let eq 'e')
    {return($nlat, 'd');}
    if ($let eq 'f')
    {return($nlat, 'e');}
    if ($let eq 'g')
    {return($nlat, 'f');}
    if ($let eq 'h')
    {return($nlat, 'g');}

    if ($let eq 'A')
    {$nlat--; return($nlat, 'H');}
    if ($let eq 'B')
    {return($nlat, 'A');}
    if ($let eq 'C')
    {return($nlat, 'B');}
    if ($let eq 'D')
    {return($nlat, 'C');}
    if ($let eq 'E')
    {return($nlat, 'D');}
    if ($let eq 'F')
    {return($nlat, 'E');}
    if ($let eq 'G')
    {return($nlat, 'F');}
    if ($let eq 'H')
    {return($nlat, 'G');}
}

sub tryNeighbor($$$$$) {
    my ($nlat, $nlon, $nlats, $nlons, $dir) = @_;
    my ($neighbor);
 
    if ($dir =~ /north/) {
	($nlat, $nlats) = nextlet($nlat, $nlats);
    }
    if ($dir =~ /west/) {
	($nlon, $nlons) = nextnum($nlon, $nlons);
    }
    if ($dir =~ /south/) {
	($nlat, $nlats) = lastlet($nlat, $nlats);
    }
    if ($dir =~ /east/) {
	($nlon, $nlons) = lastnum($nlon, $nlons);
    }

    $neighbor = "$firstchar$nlat$nlon$nlats$nlons";
	
    if (-f "$mapdir$neighbor.tif") {
	return("$dir=$neighbor.ini\n");
    }

    return("");	
}

foreach $mapdir (@dirlist)
{
    if ($mapdir !~ /\/$/ ) {
	$mapdir = "$mapdir/";
    }

    opendir (DIR, $mapdir) ||
	die "mapini: couldn't open directory \`$mapdir': $!\n";
    my ($file, $fullfile, $expr, $mlat, $mlon, $mlats, $mlons);
    my ($dlat, $dlon, $inifile, $dir);
    foreach $file (sort grep (! /^\./, readdir (DIR)))
    {
	# Only examine .tif files.
	next unless $file =~ /\.tif$/;
	next unless $file =~ /^[oO]/;
	$fullfile = $mapdir . '/' . $file;
	$inifile = $mapdir . '/' . $file;
	$inifile =~ s/\.tif$/.ini/;
	$firstchar = substr($file, 0, 1);
	open (INI, ">$inifile");
	print INI "[general]\nformat=tiff\nmaptype=drg\nfilename=$file\nmaptitle=\n";
	$file =~ /^[oO]([0-9][0-9])([0-9][0-9][0-9])([a-hA-H])([1-8])/;
	$mlat = $1;
	$mlon = $2;
	$mlats = $3;
	$mlons = $4;

	print INI "[edges]\n";
	$dlat = $mlat+(.125*(lettonum($mlats)));
	$dlon = 0-$mlon-(.125*$mlons)+.125;
	print INI "north=$dlat\n";
	print INI "east=$dlon\n";
	$dlat = $mlat-.125+(.125*(lettonum ($mlats)));
	$dlon = 0-$mlon-(.125*$mlons);
	print INI "south=$dlat\n";
	print INI "west=$dlon\n";

	print INI "[neighbors]\n";
	foreach $dir ("north", "south", "east", "west",
		      "northeast", "northwest", "southeast", "southwest") {
	    print INI tryNeighbor($mlat, $mlon, $mlats, $mlons, $dir);
	}

	print INI "[georef]\n";
	close (INI);
    }
    closedir (DIR);
}


